/*
 * Copyright (c) 2015 by inovat, innovative systeme - verkehr - tunnel - technik, Dipl.-Ing. H. C. Kniss
 *
 * This file is part of de.bsvrz.kex.tls.osi7.conversion.Fg001VerkehrsDaten
 *
 * de.bsvrz.kex.tls.osi7.conversion.Fg001VerkehrsDaten is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  de.bsvrz.kex.tls.osi7.conversion.Fg001VerkehrsDaten is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with de.bsvrz.kex.tls.osi7.conversion.Fg001VerkehrsDaten.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Diese Datei ist Teil von de.bsvrz.kex.tls.osi7.conversion.Fg001VerkehrsDaten.
 *
 *  de.bsvrz.kex.tls.osi7.conversion.Fg001VerkehrsDaten ist Freie Software: Sie knnen es unter den Bedingungen
 *  der GNU General Public License, wie von der Free Software Foundation,
 *  Version 3 der Lizenz oder (nach Ihrer Wahl) jeder spteren
 *  verffentlichten Version, weiterverbreiten und/oder modifizieren.
 *
 *  de.bsvrz.kex.tls.osi7.conversion.Fg001VerkehrsDaten wird in der Hoffnung, dass es ntzlich sein wird, aber
 *  OHNE JEDE GEWHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 *  Gewhrleistung der MARKTFHIGKEIT oder EIGNUNG FR EINEN BESTIMMTEN ZWECK.
 *  Siehe die GNU General Public License fr weitere Details.
 *
 *  Sie sollten eine Kopie der GNU General Public License zusammen mit diesem
 *  Programm erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * inovat, Dipl.-Ing. H. C. Kniss
 * An der Krautwiese 37
 * D-53783 Eitorf
 * +49 (0)2243 8464 193
 * info@inovat.de
 * www.inovat.de
 */



package de.bsvrz.kex.tls.osi7.conversion;

//~ NICHT JDK IMPORTE =========================================================

import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.kex.tls.osi7.DeBlock;
import de.bsvrz.sys.funclib.debug.Debug;

//~ KLASSEN ===================================================================

/**
 * Ermittelt die Verkehrsdaten der FG 1 der Versionen 0 bis 4 in 8 Bit und 16 Bit aus einem {@link DeBlock}.
 *
 * @author inovat, innovative systeme - verkehr - tunnel - technik
 * @author Dipl.-Ing. Hans Christian Kniss (HCK)
 */
public class Fg001VerkehrsDaten {
    private static final int NICHT_ERMITTELBAR            = -1;
    private static final int NICHT_ERMITTELBAR_FEHLERHAFT = -3;

    /** DebugLogger fr Debug-Ausgaben. */
    private static final Debug debug = Debug.getLogger();

    //~ FELDER ================================================================

    private int     _b;                                            // Belegungs-Wert aus Telegramm
    private String  _dePid;                                        // PID des DE, wird fr Betriebsmeldungen bentigt.
    private int     _qKfz;                                         // qKfz-Wert aus Telegramm
    private int     _qLkw;                                         // qLkw-Wert aus Telegramm
    private int     _qPkw;                                         // wird berechnet
    private int     _sKfz;                                         // Streuung Kfz-Wert aus Telegramm
    private int     _tNetto;                                       // tNetto-Wert aus Telegramm
    private int     _vgKfz;                                        // vgKfz-Wert aus Telegramm
    private int     _vKfz;                                         // wird berechnet
    private int     _vLkw;                                         // vLkw-Wert aus Telegramm
    private int     _vPkw;                                         // vPkw-Wert aus Telegramm
    private String  _betriebsMeldung                     = "";
    private boolean _plPruefungUndKorrekturDurchgefuehrt = false;  // Flag, damit PL nur einmal durchgefhrt wird

    //~ KONSTRUKTOREN  (und vom Konstruktor verwendete Methoden) ==============

    /**
     * Standdardkonstruktor, um Verkehrsdaten aus eine {@link DeBlock} zu erstellen.
     * <p/>
     * Beim Einlesen erfolgt direkt eine Umsetzung der TLS-Werte fr "nicht ermittelbar" in die Kennungen fr die
     * Zielattributgruppe (255, 65535 ==> -1, erkannte Fehler (z. B. bei b > 100%) wird zu -3 (nicht
     * ermittelbar/fehlerhaft)
     *
     * @param deBlock Daten eines DeBlock der FG1, Typ 49 - 53 und 113 - 117 (Version 0-4)
     */
    public Fg001VerkehrsDaten(DeBlock deBlock, String dePid) {
        _dePid = dePid;

        // Alle Werte auf "nicht ermittelbar" (-1) setzen, da nicht in allen
        // DE-Typen alle Werte enthalten sind.
        _qKfz   = NICHT_ERMITTELBAR;
        _qPkw   = NICHT_ERMITTELBAR;
        _qLkw   = NICHT_ERMITTELBAR;
        _vKfz   = NICHT_ERMITTELBAR;
        _vPkw   = NICHT_ERMITTELBAR;
        _vLkw   = NICHT_ERMITTELBAR;
        _tNetto = NICHT_ERMITTELBAR;
        _b      = NICHT_ERMITTELBAR;
        _sKfz   = NICHT_ERMITTELBAR;
        _vgKfz  = NICHT_ERMITTELBAR;

        // Die ersten vier Werte sind in allen Versionen identisch...
        switch (deBlock.getTyp()) {
        case 49 :   // Version 0 - 8-Bit-Format
        case 50 :   // Version 1 - 8-Bit-Format
        case 51 :   // Version 2 - 8-Bit-Format
        case 52 :   // Version 3 - 8-Bit-Format
        case 53 :   // Version 4 - 8-Bit-Format
            _qKfz = tls8BitToInt(deBlock.getDeData().readUnsignedByteAsInt());
            _qLkw = tls8BitToInt(deBlock.getDeData().readUnsignedByteAsInt());
            _vPkw = tls8BitToInt(deBlock.getDeData().readUnsignedByteAsInt());
            _vLkw = tls8BitToInt(deBlock.getDeData().readUnsignedByteAsInt());

            break;

        case 113 :  // Version 0 - 16-Bit-Format
        case 114 :  // Version 1 - 16-Bit-Format
        case 115 :  // Version 2 - 16-Bit-Format
        case 116 :  // Version 3 - 16-Bit-Format
        case 117 :  // Version 4 - 16-Bit-Format
            _qKfz = tls16BitToInt(deBlock.getDeData().readUnsignedLowHighByteAsInt());
            _qLkw = tls16BitToInt(deBlock.getDeData().readUnsignedLowHighByteAsInt());
            _vPkw = tls8BitToInt(deBlock.getDeData().readUnsignedByteAsInt());
            _vLkw = tls8BitToInt(deBlock.getDeData().readUnsignedByteAsInt());

            break;

        default :
            debug.warning("Unbekannter LVE-Kurzzeit-Typ (49/113 - 53/117): " + deBlock.getTyp());

            return;
        }

        // ...und jetzt noch weitere Werte abhngig von der Version einlesen
        switch (deBlock.getTyp()) {
        case 49 :
        case 113 :
            break;

        case 50 :                // Version 1 - 8-Bit-Format
        case 114 :               // Version 1 - 16-Bit-Format
            _tNetto = tls8BitToInt(deBlock.getDeData().readUnsignedByteAsInt());

            if (_tNetto > 0) {   // Falls keine Fehlerkennung vorliegt...
                _tNetto *= 100;  // ...Umrechnung in Millisekunden
            }

            break;

        case 51 :                // Version 2 - 8-Bit-Format
        case 115 :               // Version 2 - 16-Bit-Format
            _b = tls8BitToInt(deBlock.getDeData().readUnsignedByteAsInt());

            if (_b > 100) {
                _b = NICHT_ERMITTELBAR_FEHLERHAFT;
            }

            break;

        case 52 :                // Version 3 - 8-Bit-Format
        case 116 :               // Version 3 - 16-Bit-Format
        case 53 :                // Version 4 - 8-Bit-Format
        case 117 :               // Version 4 - 16-Bit-Format
            _tNetto = tls8BitToInt(deBlock.getDeData().readUnsignedByteAsInt());

            if (_tNetto > 0) {   // Falls keine Fehlerkennung vorliegt...
                _tNetto *= 100;  // ...Umrechnung in Millisekunden
            }

            _b = tls8BitToInt(deBlock.getDeData().readUnsignedByteAsInt());

            if (_b > 100) {
                _b = NICHT_ERMITTELBAR_FEHLERHAFT;
            }

            _sKfz  = tls8BitToInt(deBlock.getDeData().readUnsignedByteAsInt());
            _vgKfz = tls8BitToInt(deBlock.getDeData().readUnsignedByteAsInt());

            break;

        default :
            debug.warning("Unbekannter LVE-Kurzzeit-Typ (49/113 - 53/117): " + deBlock.getTyp());
        }
    }

    /**
     * Konstruktor fr JUnitTest, um Testdaten unter Umgehung von DeBloecken zu erzeugen
     *
     * @param qKfz Sollvorgabe fr qKfz
     * @param qPkw Sollvorgabe fr qPkw
     * @param qLkw Sollvorgabe fr qLkw y
     * @param vKfz Sollvorgabe fr vKfz
     * @param vPkw Sollvorgabe fr vPkw
     * @param vLkw Sollvorgabe fr vLkw
     */
    public Fg001VerkehrsDaten(int qKfz, int qPkw, int qLkw, int vKfz, int vPkw, int vLkw) {
        _dePid  = "DE.fuer.JUnit.Tests";
        _qKfz   = qKfz;
        _qPkw   = qPkw;
        _qLkw   = qLkw;
        _vKfz   = vKfz;
        _vPkw   = vPkw;
        _vLkw   = vLkw;
        _tNetto = -1;
        _b      = -1;
        _sKfz   = -1;
        _vgKfz  = -1;
    }

    //~ METHODEN ==============================================================

    @Override
    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }

        if ((o == null) || (getClass() != o.getClass())) {
            return false;
        }

        Fg001VerkehrsDaten that = (Fg001VerkehrsDaten) o;

        if (_b != that._b) {
            return false;
        }

        if (_qKfz != that._qKfz) {
            return false;
        }

        if (_qLkw != that._qLkw) {
            return false;
        }

        if (_qPkw != that._qPkw) {
            return false;
        }

        if (_sKfz != that._sKfz) {
            return false;
        }

        if (_tNetto != that._tNetto) {
            return false;
        }

        if (_vgKfz != that._vgKfz) {
            return false;
        }

        if (_vKfz != that._vKfz) {
            return false;
        }

        if (_vLkw != that._vLkw) {
            return false;
        }

        if (_vPkw != that._vPkw) {
            return false;
        }

        return true;
    }

    public Data ermittleDatenAtgVerkehrsDatenKurzZeitIntervallAspExterneErfassung(Data data) {

        // Werte jeweils auf Zustand "nicht ermittelbar" setzen.
        // PL-Flags etc. vorbesetzen
        String[] valStrings = { "qKfz", "vKfz", "qLkw", "vLkw", "qPkw", "vPkw", "b", "tNetto", "sKfz", "vgKfz" };

        for (String valString : valStrings) {
            data.getItem(valString).getUnscaledValue("Wert").setText("nicht ermittelbar");
            data.getItem(valString).getItem("Status").getItem("Erfassung").getUnscaledValue("NichtErfasst").setText("Nein");
            data.getItem(valString).getItem("Status").getItem("PlFormal").getUnscaledValue("WertMax").setText("Nein");
            data.getItem(valString).getItem("Status").getItem("PlFormal").getUnscaledValue("WertMin").setText("Nein");
            data.getItem(valString).getItem("Status").getItem("PlLogisch").getUnscaledValue("WertMaxLogisch").setText("Nein");
            data.getItem(valString).getItem("Status").getItem("PlLogisch").getUnscaledValue("WertMinLogisch").setText("Nein");
            data.getItem(valString).getItem("Status").getItem("MessWertErsetzung").getUnscaledValue("Implausibel").setText("Nein");
            data.getItem(valString).getItem("Status").getItem("MessWertErsetzung").getUnscaledValue("Interpoliert").setText("Nein");
            data.getItem(valString).getItem("Gte").getUnscaledValue("Index").set(-1);
            data.getItem(valString).getItem("Gte").getUnscaledValue("Verfahren").set(0);
        }

        // qKfz
        data.getItem("qKfz").getUnscaledValue("Wert").set(_qKfz);
        data.getItem("qKfz").getItem("Gte").getScaledValue("Index").set(1.0);

        // qPkw - Wert wurde berechnet
        data.getItem("qPkw").getUnscaledValue("Wert").set(_qPkw);
        data.getItem("qPkw").getItem("Gte").getScaledValue("Index").set(1.0);
        data.getItem("qPkw").getItem("Status").getItem("Erfassung").getUnscaledValue("NichtErfasst").setText("Ja");

        // qLkw
        data.getItem("qLkw").getUnscaledValue("Wert").set(_qLkw);
        data.getItem("qLkw").getItem("Gte").getScaledValue("Index").set(1.0);

        // vPkw
        data.getItem("vPkw").getUnscaledValue("Wert").set(_vPkw);
        data.getItem("vPkw").getItem("Gte").getScaledValue("Index").set(1.0);

        // vLkw
        data.getItem("vLkw").getUnscaledValue("Wert").set(_vLkw);
        data.getItem("vLkw").getItem("Gte").getScaledValue("Index").set(1.0);

        // vKfz - Wert wurde berechnet
        data.getItem("vKfz").getUnscaledValue("Wert").set(_vKfz);
        data.getItem("vKfz").getItem("Gte").getScaledValue("Index").set(1.0);
        data.getItem("vKfz").getItem("Status").getItem("Erfassung").getUnscaledValue("NichtErfasst").setText("Ja");

        // tNetto
        data.getItem("tNetto").getUnscaledValue("Wert").set(_tNetto);
        data.getItem("tNetto").getItem("Gte").getScaledValue("Index").set(1.0);

        // b
        data.getItem("b").getUnscaledValue("Wert").set(_b);
        data.getItem("b").getItem("Gte").getScaledValue("Index").set(1.0);

        // sKfz
        data.getItem("sKfz").getUnscaledValue("Wert").set(_sKfz);
        data.getItem("sKfz").getItem("Gte").getScaledValue("Index").set(1.0);

        // vgKfz
        data.getItem("vgKfz").getUnscaledValue("Wert").set(_vgKfz);
        data.getItem("vgKfz").getItem("Gte").getScaledValue("Index").set(1.0);

        return data;
    }

    //~ GET METHODEN ==========================================================

    public int getB() {
        return _b;
    }

    public String getBetriebsMeldung() {
        return _betriebsMeldung;
    }

    //~ METHODEN ==============================================================

    public int getqKfz() {
        return _qKfz;
    }

    public int getqLkw() {
        return _qLkw;
    }

    public int getqPkw() {
        return _qPkw;
    }

    public int getsKfz() {
        return _sKfz;
    }

    public int gettNetto() {
        return _tNetto;
    }

    //~ GET METHODEN ==========================================================

    public int getVgKfz() {
        return _vgKfz;
    }

    //~ METHODEN ==============================================================

    public int getvKfz() {
        return _vKfz;
    }

    public int getvLkw() {
        return _vLkw;
    }

    public int getvPkw() {
        return _vPkw;
    }

    @Override
    public int hashCode() {
        int result = _b;

        result = 31 * result + _qKfz;
        result = 31 * result + _qLkw;
        result = 31 * result + _qPkw;
        result = 31 * result + _sKfz;
        result = 31 * result + _tNetto;
        result = 31 * result + _vgKfz;
        result = 31 * result + _vKfz;
        result = 31 * result + _vLkw;
        result = 31 * result + _vPkw;

        return result;
    }

    //~ GET METHODEN ==========================================================

    public boolean isPlPruefungUndKorrekturDurchgefuehrt() {
        return _plPruefungUndKorrekturDurchgefuehrt;
    }

    //~ METHODEN ==============================================================

    /**
     * Fhrt die PLPrfung und Korrektur der empfangenen Daten gem AFo DUA, Anforderung KEx-15 durch
     */
    protected void plPruefungUndKorrektur() {

        // PL-Prfung und Korrektur darf nur einmal durchgefhrt werden
        if (isPlPruefungUndKorrekturDurchgefuehrt()) {
            return;
        }

        // PL-Prfung und Korrektur gemaess AFo DUA KEx-15 durchfhren
        _plPruefungUndKorrekturDurchgefuehrt = true;

        // Bedingungen/Folgen:
        // B1, B2, F1
        if ((_qKfz >= 0) && (_qLkw >= 0) && (_qKfz >= _qLkw)) {
            _qPkw = _qKfz - _qLkw;
        } else {

            // Betriebsmeldung erstellen
            _betriebsMeldung = String.format("Attribute q nicht ermittelbar/fehlerhaft an DE [%s], da qKfz=%d und qLkw=%d. ", _dePid, _qKfz, _qLkw);
            _qKfz            = NICHT_ERMITTELBAR_FEHLERHAFT;
            _qPkw            = NICHT_ERMITTELBAR_FEHLERHAFT;
            _qLkw            = NICHT_ERMITTELBAR_FEHLERHAFT;
        }

        // Bedingungen/Folgen:
        // B3, B4, B5, B6, F2
        if ((_qKfz > 0) && (_qPkw >= 0) && (_qLkw >= 0) && (_vPkw >= 0) && (_vLkw >= 0)) {
            _vKfz = (int) ((((_qPkw * _vPkw) + (_qLkw * _vLkw)) / (double) _qKfz) + 0.5);

            return;  // B3 erfllt, ok und fertig
        }

        if ((_qPkw == 0) && (_vPkw == NICHT_ERMITTELBAR)) {
            _vKfz = _vLkw;

            return;  // B4 erfllt, ok und fertig
        }

        if ((_qLkw == 0) && (_vLkw == NICHT_ERMITTELBAR)) {
            _vKfz = _vPkw;

            return;  // B5 erfllt, ok und fertig
        }

        // Sonst B6 mit F2
        // Betriebsmeldung erstellen
        if (_betriebsMeldung.equals("")) {
            _betriebsMeldung = String.format("Attribute v nicht ermittelbar/fehlerhaft an DE [%s], da qKfz=%d, qPkw=%d, qLkw=%d, vPkw=%d und vLkw=%d.", _dePid, _qKfz, _qPkw, _qLkw, _vPkw, _vLkw);
        } else {
            _betriebsMeldung += String.format("Attribute v nicht ermittelbar/fehlerhaft an DE [%s], da qKfz=%d, qPkw=%d, qLkw=%d, vPkw=%d und vLkw=%d.", _dePid, _qKfz, _qPkw, _qLkw, _vPkw, _vLkw);
        }

        _vKfz = NICHT_ERMITTELBAR_FEHLERHAFT;
        _vPkw = NICHT_ERMITTELBAR_FEHLERHAFT;
        _vLkw = NICHT_ERMITTELBAR_FEHLERHAFT;
    }

    /**
     * Prft, ob der bergebene Wert == 0xffff ist. In diesem Fall wird -1 zurckgegeben, so dass dieser als
     * unscaledValue direkt gesetzt werden kann (entspricht dem Zustand "nicht ermittelbar").
     *
     * @param value Der zu prfende Wert (der als 16 Bit vorzeichenlos interpretiert wird)
     *
     * @return value, wenn ungleich 0xffff, sonst -1.
     */
    private int tls16BitToInt(int value) {
        if (value == 0xffff) {
            return -1;
        } else {
            return value;
        }
    }

    /**
     * Prft, ob der bergebene Wert == 0xff ist. In diesem Fall wird -1 zurckgegeben, so dass dieser als unscaledValue
     * direkt gesetzt werden kann (entspricht dem Zustand "nicht ermittelbar").
     *
     * @param value Der zu prfende Wert (der als 8 Bit vorzeichenlos interpretiert wird)
     *
     * @return value, wenn ungleich 255, sonst -1.
     */
    private int tls8BitToInt(int value) {
        if (value == 0xff) {
            return -1;
        } else {
            return value;
        }
    }

    @Override
    public String toString() {
        return String.format("Fg001VerkehrsDaten: qKfz=%3d, qPkw=%3d, qLkw=%3d, vKfz=%3d, vPkw=%3d, vLkw=%3d, tNetto=%3d, b=%3d, sKfz=%3d, vgKfz=%3d", _qKfz, _qPkw, _qLkw, _vKfz, _vPkw, _vLkw, _tNetto, _b, _sKfz, _vgKfz);
    }
}


//~Formatiert mit 'inovat Kodierkonvention' am 05.08.15
